#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <voxel_loop.h>
#include <time_stamp.h>
#include <minc_def.h>

#ifndef TRUE
#  define TRUE 1
#endif
#ifndef FALSE
#  define FALSE 0
#endif

void voxel_func(void *caller_data, long num_voxels,
                int input_num_buffers, int input_vector_length,
                double *input_data[],
                int output_num_buffers, int output_vector_length,
                double *output_data[],
                Loop_Info *loop_info);

void input_func(void *caller_data, int input_mincid, int input_curfile,
                Loop_Info *loop_info);

void output_func(void *caller_data, int output_mincid, int output_curfile,
                 Loop_Info *loop_info);

void start_func(void *caller_data, long num_voxels, int output_num_buffers,
                int output_vector_length, double *output_data[],
                Loop_Info *loop_info);

void finish_func(void *caller_data, long num_voxels, int output_num_buffers,
                 int output_vector_length, double *output_data[],
                 Loop_Info *loop_info);

int main(int argc, char *argv[])
{
   int test_value = 42;
   int num_input_files, num_output_files;
   char **input_files, **output_files;
   char *arg_string;
   Loop_Options *loop_options = NULL;

   arg_string = time_stamp(argc, argv);

   if (argc > 4) {
      num_input_files = argc - 3;
      input_files = &argv[1];
      num_output_files = 2;
      output_files = &argv[argc-2];
   }
   else if (argc > 2) {
      num_input_files = argc - 2;
      input_files = &argv[1];
      num_output_files = 1;
      output_files = &argv[argc-1];
   }
   else if (argc > 1) {
      num_input_files = argc - 1;
      input_files = &argv[1];
      num_output_files = 0;
      output_files = NULL;
   }
   else {
      (void) fprintf(stderr, "Bad args.\n");
      exit(EXIT_FAILURE);
   }
#if 0
   num_input_files = argc - 1;
   input_files = &argv[1];
   num_output_files = 0;
   output_files = NULL;
#endif
   

   loop_options = create_loop_options();
   set_loop_verbose(loop_options, FALSE);
   set_loop_input_file_function(loop_options, input_func);
   set_loop_check_dim_info(loop_options, FALSE);
   set_loop_accumulate(loop_options, TRUE, 0, NULL, NULL);
#if 0
   set_loop_dimension(loop_options, MIzspace);
   set_loop_clobber(loop_options, TRUE);
   set_loop_datatype(loop_options, NC_SHORT, TRUE, 0.0, 32000.0);
   set_loop_convert_input_to_scalar(loop_options, FALSE);
   set_loop_output_vector_size(loop_options, 0);
   set_loop_copy_all_header(loop_options, TRUE);
   set_loop_output_file_function(loop_options, output_func);
#define NOACCUMULATE 0
#if NOACCUMULATE
   set_loop_accumulate(loop_options, FALSE, 0, start_func, finish_func);
#else
   set_loop_accumulate(loop_options, TRUE, 0, start_func, finish_func);
#endif
   set_loop_buffer_size(loop_options, 2*1024*1024);
#endif

   voxel_loop(num_input_files, input_files, num_output_files, output_files,
              arg_string, loop_options, voxel_func, (void *) &test_value);

   free_loop_options(loop_options);

   return EXIT_SUCCESS;
}

void print_loop_info(Loop_Info *loop_info)
{
   long start[MAX_VAR_DIMS], count[MAX_VAR_DIMS];
   int idim;
   
   get_info_shape(loop_info, MAX_VAR_DIMS, start, count);
   (void) printf("Start =");
   for (idim=0; idim < 10; idim++) (void) printf(" %d", (int) start[idim]);
   (void) printf("\n");
   (void) printf("Count =");
   for (idim=0; idim < 10; idim++) (void) printf(" %d", (int) count[idim]);
   (void) printf("\n");

   (void) printf("Current file = %d\n", get_info_current_file(loop_info));
   (void) printf("Current index = %d\n", get_info_current_index(loop_info));

}

void voxel_func(void *caller_data, long num_voxels,
                int input_num_buffers, int input_vector_length,
                double *input_data[],
                int output_num_buffers, int output_vector_length,
                double *output_data[],
                Loop_Info *loop_info)
{
   int *testvalue;
   int ivox, ifile, ofile, ivec, ovec;
   double value, invalue;

   testvalue = (int *) caller_data;
   (void) printf("Voxel function\n");
   (void) printf("Caller data = %d\n", *testvalue);
   (void) printf("Num voxels  = %d\n", (int) num_voxels);
   (void) printf("Input buffers = %d\n", input_num_buffers);
   (void) printf("Input vector length = %d\n", input_vector_length);
   (void) printf("Output buffers = %d\n", output_num_buffers);
   (void) printf("Output vector length = %d\n", output_vector_length);
   print_loop_info(loop_info);
   (void) printf("\n");

#if 1
   for (ifile=0; ifile < input_num_buffers; ifile++) {
      for (ivox=0; ivox < num_voxels; ivox++) {
         for (ivec=0; ivec < input_vector_length; ivec++) {
            (void) printf(" %g", 
                          input_data[ifile][ivox*input_vector_length+ivec]);
         }
         (void) printf("\n");
      }
   }
#endif

   for (ivox=0; ivox < num_voxels; ivox++) {
      value = 0;
      for (ifile=0; ifile < input_num_buffers; ifile++) {
         for (ivec=0; ivec < input_vector_length; ivec++) {
            invalue = input_data[ifile][ivox*input_vector_length+ivec]; 
            if (invalue != -DBL_MAX)
               value += invalue;
         }
      }
      for (ofile=0; ofile < output_num_buffers; ofile++) {
         for (ovec=0; ovec < output_vector_length; ovec++) {
#if NOACCUMULATE
            output_data[ofile][ivox*output_vector_length+ovec] = 
               value + 100*ofile + ovec;
#else
            output_data[ofile][ivox*output_vector_length+ovec] += 
               value + 100*ofile + ovec;
#endif
         }
      }
   }

   (void) printf("\n\n");

}

void input_func(void *caller_data, int input_mincid, int input_curfile,
                Loop_Info *loop_info)
{
   int *testvalue;
   char string[1024];

   testvalue = (int *) caller_data;
   (void) printf("Input function\n");
   (void) printf("Caller data = %d\n", *testvalue);
   (void) printf("Current file = %d\n", input_curfile);
   (void) miattgetstr(input_mincid, NC_GLOBAL, MIhistory, sizeof(string),
                      string);
   (void) printf("History :\n%s", string);
   print_loop_info(loop_info);
   (void) printf("\n");
}

void output_func(void *caller_data, int output_mincid, int output_curfile,
                 Loop_Info *loop_info)
{
   int *testvalue;
   char string[1024];

   testvalue = (int *) caller_data;
   (void) printf("Output function\n");
   (void) printf("Caller data = %d\n", *testvalue);
   (void) printf("Current file = %d\n", output_curfile);
   (void) sprintf(string, "This is a test string for file %d", output_curfile);
   (void) miattputstr(output_mincid, NC_GLOBAL, "teststring", string);
   print_loop_info(loop_info);
   (void) printf("\n");
}

void start_func(void *caller_data, long num_voxels, int output_num_buffers,
                int output_vector_length, double *output_data[],
                Loop_Info *loop_info)
{
   int *testvalue;
   int ibuff, ivox;

   testvalue = (int *) caller_data;
   (void) printf("Start function\n");
   (void) printf("Caller data = %d\n", *testvalue);
   (void) printf("Num voxels  = %d\n", (int) num_voxels);
   (void) printf("Output buffers = %d\n", output_num_buffers);
   (void) printf("Output vector length = %d\n", output_vector_length);
   print_loop_info(loop_info);
   (void) printf("\n");

   for (ibuff=0; ibuff < output_num_buffers; ibuff++) {
      for (ivox=0; ivox < num_voxels*output_vector_length; ivox++) {
         output_data[ibuff][ivox] = -20 * (ibuff+1);
      }
   }

}

void finish_func(void *caller_data, long num_voxels, int output_num_buffers,
                 int output_vector_length, double *output_data[],
                 Loop_Info *loop_info)
{
   int *testvalue;
   int ibuff, ivox;

   testvalue = (int *) caller_data;
   (void) printf("Finish function\n");
   (void) printf("Caller data = %d\n", *testvalue);
   (void) printf("Num voxels  = %d\n", (int) num_voxels);
   (void) printf("Output buffers = %d\n", output_num_buffers);
   (void) printf("Output vector length = %d\n", output_vector_length);
   print_loop_info(loop_info);
   (void) printf("\n");

   for (ibuff=0; ibuff < output_num_buffers; ibuff++) {
      for (ivox=0; ivox < num_voxels*output_vector_length; ivox++) {
         output_data[ibuff][ivox] /= (10 * (ibuff+1));
      }
   }
}

